
// For easier reading, do: Menubar / Editor / Show Rendered Markup


/*:
## Demo 1
### Inheritance Good thing 1: _implementation reuse_
*/


import Foundation

class Animal1 {
  func behave() {
    print("I am an instance of type: \(self)")
    print("I wake up in the morning")
    print("I make my animal sounds")
    print("I go to sleep at night")
    print("")
  }
}

class Dog1: Animal1 { }

//:  Every instance of `Dog` automatically behaves like an animal, just by _being_ an `Animal`

let dog1 = Dog1()
dog1.behave()



//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:
### Inheritance Good Thing 2: overriding provides a _mechanism for customization_
*/

// example here: override just one method in defining a new animal

class Animal2 {
  // the superclass provides functionality that subclasses rely on inheriting.
  func behave() {
    print("I am an instance of type: \(self)")
    print("I wake up in the morning")
    self.speak()
    print("I go to sleep at night")
    print("")
  }
  
  func speak() {
    print("I am making some undefined kind of animal sound")
  }
}

class Dog2: Animal2 {
  override func speak() {
    print("I go woof")
  }
}

let dog2 = Dog2()
dog2.behave()

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////


/*:
### Inheritance Good Thing 3: subclassing provides _interface reuse_ ...
*/

class Animal3 {
  // the superclass provides functionality that subclasses rely on inheriting.
  func behave() {
    print("I am an instance of type: \(self)")
    print("I wake up in the morning")
    self.speak()
    print("I go to sleep at night")
    print("")
  }
  
  func speak() {
    print("I am making some undefined kind of animal sound")
  }
}

class Dog3: Animal3 {
  override func speak() {
    print("I go woof")
  }
}

let dog3 = Dog3()
dog3.behave()

class Cat3: Animal3 {
  override func speak() {
    print("I go meow")
  }
}

let cat3 = Cat3()
cat3.behave()

/*:
### ... and interface re-use enables modularity
*/

let mixedAnimals3: [Animal3] = [Dog3(), Cat3(), Cat3()]

func runBarnyardWithResidents(residents: [Animal3]) {
  print("Executing: \(#function)")
  for animal in residents {
    animal.behave()
  }
}

runBarnyardWithResidents(mixedAnimals3)

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
/*:
### Inheritance Good Thing 4: Subclasses provide _dynamic_ dispatch
*/

// returns a random mixed array of dogs and cats
func randomAnimals() -> [Animal3] {
  var result: [Animal3] = []
  for _ in 0..<5 {
    let coinFlip = arc4random_uniform(2)
    let animal = coinFlip == 0 ? Dog3() : Cat3()
    result.append(animal)
  }
  return result
}
print("=== start barnyard of random animals")
runBarnyardWithResidents(randomAnimals())
print("=== end barnyard of random animals")

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:
### Inheritance Bad Thing 1: You cannot use superclasses or inheritance with Swift value types.
*/

// example here briefly showing errors when you try to define a Swift subclass

class Animal4 {
  // the superclass provides functionality that subclasses rely on inheriting.
  func behave() {
    print("I wake up in the morning")
    self.speak()
    print("I go to sleep at night")
  }
  
  func speak() {
    print("I am making some kind of animal sound")
  }
}

// // uncomment to try:
// struct Cow: Animal4 { }
/*:

When we try to do:

`struct Cow: Animal4 { }`

The compiler says:
`Non-class type 'Cow" cannot inherit from class 'Animal4'`

This is a *major limitation* on value types (i.e., structs and enums), or on superclasses, depending on how you look at.

*/


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### Inheritance Bad Thing 2: as a means for customization, inheritance is often burdensome and imprecise

*/

class Animal5 {
  let name: String
  init(name: String) { self.name = name }
  
  // the superclass provides functionality that subclasses rely on inheriting.
  final func behave() {
    print("\(self.name) wakes up in the morning")
    self.speak()
    print("I go to sleep at night")
  }
  
  // the superclass provides a function that all subclasses MUST override
  func speak() {
    print("I am making some undefined kind of animal sound")
  }
}

//: We don't want this to be possible. Every instance should be some specific type of animal (`Animal5` should be what's called an abstract class in other laguages)
let animal = Animal5(name: "Romeo")

//: We don't want this to be possible. We want every Animal5 subclass to override `speak` in its own way. (`speak` should be what's called an abstract method in other languages.)
class BrokenRooster: Animal5 {
}

//: And are we sure we are initializing correctly? How do we _preserve superclass invariants_? To take a somewhat unlikely example, what if as a matter of its internal implementation, `Animal5` only provided enough storage for four character names? How does the subclass know about this and enforce this invariant? A more realistic example: if a class is initialized with a URL, how do you know which URL schemes it supports? Does your subclass need to support all of them to fulfill the is-a requirement?

let wronglyNamedRooster = BrokenRooster(name: "Totoro")

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### Inheritance Bad Thing 3: Upfront modelling (not retroactive modelling)

*/

class Animal6 {
  func behave() { print("I'm waking up, making noise, going to sleep") }
}
class Dog6: Animal6 { }
class Cat6: Animal6 { }

class Vehicle6 {
  func convey() { print("I'm carrying you from one place to another") }
}
class Car6: Vehicle6 { }
class Bicycle6: Vehicle6 { }

//: But where do we put `Horse`, which is both an animal and a vehicle? If we start by defining it as an Animal, then we cannot later also make it a Vehicle. Or vice versa.

// This horse automatically wakes up etc., but it doesn't look like a `Vehicle` and it doesn't convey you anywhere
class Horse1a: Animal6 { }

// This horse automatically conveys you, but it doesn't look like an `Animal` and it doesn't wake up etc..
class Horse1b: Vehicle6 { }

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

## Demo 2: How protocols replace and improve on superclasses

### Protocol Good Thing: Protocols allow _pure interface re-use_ (unlike superclasses)

*/

protocol Animal {
  func behave()
  func speak()
}

protocol Vehicle {
  func convey()
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### Protocol Good Thing: protocol extensions allow _more precise implementation re-use_

*/

extension Animal {
  func behave() {
    print("I wake up in the morning")
    self.speak()
    print("I go to sleep at night")
  }
}

extension Vehicle {
  func convey() { print("I'm carrying you from one place to another") }
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### Protocol Good Thing: Protocol adopters get dynamic dispatch, like subclasses

*/

struct Dog { }
struct Cat { }

//: Protocol-adopting extensions let us
extension Dog: Animal {
  func speak() { print("Woof") }
}
extension Cat: Animal {
  func speak() { print("Meow") }
}

func runBarnyardWithResidents2(residents: [Animal]) {
  for animal in residents {
    animal.behave()
    print("")
  }
}

let variousAnimals: [Animal] = [Dog(), Dog(), Cat(), Dog()]
runBarnyardWithResidents2(variousAnimals)

/*:

### (Protocol + protocol extension) ≈ Superclass

*/


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### Protocol Good Thing: Protocols allows retroactive, ad-hoc modelling

*/

struct Car { }
extension Car: Vehicle {
  func convey() { print("I make a loud vroom when I go") }
}

struct Bicycle : Vehicle {
  func convey() { print("I ring a little bell when I go") }
}

struct Horse { }

//: This answers the question of whether a horse is a vehicle or an animal. It is both!

extension Horse: Vehicle {
  func convey() { print("I carry you on a saddle on my back") }
}

extension Horse: Animal {
  func speak() { print("I winny and neigh") }
}

let horse = Horse()
horse.speak()
horse.convey()

//: Horse participates in dynamic dispatch as a kind of vehicle

func runBusyStreet(vehicles: [Vehicle]) {
  for vehicle in vehicles {
    vehicle.convey()
  }
}

runBusyStreet([Horse(),Bicycle(),Car()])


/*:

Key points from Parts 1 & 2:

Protocols + protocol extensions do the _same job_ as superclasses. That job is:
- implementation reuse
- provide points for customization
- interface reuse
- supporting modular design via dynamic dispatch on reused interfaces

But, protocols + protocol extensions are also _strictly superior_ to superclasses in a number of ways without significant tradeoffs:
- can be used with value types
- separation of interface and implementation
- more granular specification of customization points
- less monolithic: types can get the interface and implementation from multiple protocols simultaneously
- less rigid: you can retroactively add or modify default behavior, or customized behavior, or protocol participation

This is all before we come to the last topic, protocols with associated types

*/

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

## Part 2, Demo 3: Protocols with associated types (PATs)

### Inheritance Bad Thing: Superclasses/subclasses cannot express rich type relationships

*/

class Food8 { }
class Grass8: Food8 { }
class Seed8: Food8 { }

class Animal8 {
  func eat(food: Food8) { }
}

/*:

This attempt to define Cow as a Grass-eater
*/
// // uncomment to try:
//class Cow8: Animal8 {
//  override func eat(food: Grass8) {
//    print("I eat only grass, not food in general. Is that a problem??")
//  }
//}
/*:

Generates this error:

> error: Method does not override any method from its superclass

Because that definition would violate the "is-a" substitutability relation, which requires that we be able to write something like:
*/
// // would not run with above definition:
// let c: Animal8 = Cow8()
// let f: Food8 = Seed8()
// c.eat(f)
/*:

This is a problem both with superclass/subclasses, and with ordinary protocols and their adopting types.

That is because this is due to fundamental limits of the subtype/supertype relationship, which is between interfaces. It has nothing to do with how or whether you inherit implementation.


*/

protocol Food9 { }
class Grass9: Food9 { }

protocol Animal9 {
  func eat(food: Food9)
}

/*:

So this also fails:

*/
// // uncomment to try:
//class Cow9: Animal9 {
//  override func eat(food: Grass9) { }
//}
/*:

*/

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### PAT Good Thing: Protocols with Associated Types can capture richer type relationships

*/

protocol Food10 { }
class Grass10: Food10 { }

protocol PATAnimal10 {
  associatedtype SuitableFood  // declares some associated type is required of adopters
  func eat(food: SuitableFood)
}

struct Cow10 { }
extension Cow10: PATAnimal10 {
  typealias SuitableFood = Grass10 // declares that for Cow, that type is Grass
  func eat(food: Grass10) {
    print("I eat only grass")
  }
}

/*:

Swift can also infer that `typealias SuitableFood = Grass10` from your method, so you can omit that line.

*/


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### PAT Bad Thing: they cannot be used to define variables, only as constraints

*/


/*:

So this is illegal:

> let animal: PATAnimal10 = Cow10()
*/

// // uncomment to try:
// let animal: PATAnimal10 = Cow10()

/*:
If you try, you get the notorious error:

> error: `Protocol 'PATAnimal10' can only be used as a generic constraint because it has Self or associated type requirements`

*/

extension PATAnimal10 {
  func speak() { print("I only eat \(SuitableFood.self)") }
}


func runBarnyardWithResidentsGenerically<A: PATAnimal10>(residents: [A]) {
  for animal in residents {
    animal.speak()
    print("")
  }
}

runBarnyardWithResidentsGenerically([Cow10()])

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*:

### PAT Bad Thing: PATs do not support dynamic dispatch

*/

func runBarnyardWithResidentsSpecialized(residents: [Cow10]) { }
